import sharp from 'sharp';
import uid from 'uid-safe';

/******************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

function __generator(thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
}

function __read(o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
}

function __spreadArray(to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
}

var createRandomInteger = function (min, max) {
    if (min === void 0) { min = 0; }
    if (max === void 0) { max = 2147483646; }
    return Math.floor(Math.random() * (max - min + 1) + min);
};
var randShade = function (hue) { return ({
    hue: hue,
    saturation: createRandomInteger(50, 70),
    lightness: createRandomInteger(50, 60),
}); };
var hslString = function (color) {
    return "hsl(".concat(color.hue, ", ").concat(color.saturation, "%, ").concat(color.lightness, "%)");
};
var randScheme = function (base) {
    return [
        randShade(base),
        randShade((base + 60) % 360),
        randShade((base - 30) % 360),
        randShade((base + 30) % 360),
        randShade((base - 60) % 360),
    ].map(function (color) { return hslString(color); });
};
var svgRect = function (x, y, gridWidth, gridHeight, color) {
    return "<rect filter=\"url(#noise)\" x=\"".concat(x, "\" y=\"").concat(y, "\" width=\"").concat(gridWidth, "\" height=\"").concat(gridHeight, "\" fill=\"").concat(color, "\"/>");
};
var svgGridPattern = function (width, height, gridWidth, gridHeight, scheme) {
    return __spreadArray([], __read(Array(Math.floor(height / gridHeight)).keys()), false).map(function (y) {
        return __spreadArray([], __read(Array(Math.floor(width / gridWidth)).keys()), false).map(function (x) {
            return svgRect(x * gridWidth, y * gridHeight, gridWidth, gridHeight, scheme[x % 2]);
        });
    })
        .flat()
        .join("");
};
var backgroundSvg = function (width, height, _a) {
    var _b = _a === void 0 ? {} : _a, _c = _b.gridWidth, gridWidth = _c === void 0 ? createRandomInteger(5, 50) : _c, _d = _b.gridHeight, gridHeight = _d === void 0 ? createRandomInteger(5, 50) : _d, _e = _b.scheme, scheme = _e === void 0 ? randScheme(createRandomInteger(0, 360)) : _e, _f = _b.seeds, seeds = _f === void 0 ? [createRandomInteger(), createRandomInteger()] : _f;
    return "<svg width=\"".concat(width, "\" height=\"").concat(height, "\" viewBox=\"0 0 ").concat(width, " ").concat(height, "\"><filter id=\"noise\"><feTurbulence type=\"turbulence\" baseFrequency=\"0.005\" seed=\"").concat(seeds[0], "\" numOctaves=\"2\" result=\"turbulence\"/><feDisplacementMap in2=\"turbulence\" in=\"SourceGraphic\" scale=\"30\" xChannelSelector=\"R\" yChannelSelector=\"G\"/></filter><filter id=\"heavy\"><feTurbulence type=\"turbulence\" baseFrequency=\"0.005\" seed=\"").concat(seeds[1], "\" numOctaves=\"2\" result=\"turbulence\"/><feDisplacementMap in2=\"turbulence\" in=\"SourceGraphic\" scale=\"100\" xChannelSelector=\"R\" yChannelSelector=\"G\"/></filter><rect width=\"").concat(width, "\" height=\"").concat(height, "\" fill=\"").concat(scheme[4], "\"/><rect filter=\"url(#heavy)\"  width=\"").concat(width / 2, "\" height=\"").concat(height, "\" x=\"").concat(width / 5, "\" fill=\"").concat(scheme[2], "\"/><rect filter=\"url(#heavy)\" width=\"").concat(width / 2, "\" height=\"").concat(height, "\" x=\"").concat(width / 2, "\" fill=\"").concat(scheme[3], "\"/>").concat(svgGridPattern(width, height, gridWidth, gridHeight, scheme), "</svg>");
}; // eslint-disable-line
// prettier-ignore
var puzzlePieceSvg = function (_a) {
    var _b = _a === void 0 ? {} : _a, _c = _b.distort, distort = _c === void 0 ? false : _c, _d = _b.rotate, rotate = _d === void 0 ? false : _d, _e = _b.fill, fill = _e === void 0 ? '#000' : _e, _f = _b.stroke, stroke = _f === void 0 ? '#fff' : _f, _g = _b.seed, seed = _g === void 0 ? 0 : _g, _h = _b.opacity, opacity = _h === void 0 ? '0.5' : _h, _j = _b.strokeWidth, strokeWidth = _j === void 0 ? '0.25' : _j;
    return "<svg viewBox=\"0 0 20 20\" height=\"60\" width=\"60\"><filter id=\"noise\"><feTurbulence type=\"turbulence\" baseFrequency=\"0.05\" seed=\"".concat(seed, "\" numOctaves=\"2\" result=\"turbulence\"/><feDisplacementMap in2=\"turbulence\" in=\"SourceGraphic\" scale=\"2.5\" xChannelSelector=\"R\" yChannelSelector=\"G\"/></filter><path ").concat(distort ? 'filter="url(#noise)"' : '', " ").concat(rotate ? "transform=\"rotate(".concat(seed, ", 10, 10)\"") : '', " d=\"M5.56.56a2.305 2.305 0 00-2.296 2.304 2.305 2.305 0 00.801 1.747H.135v4.295a2.305 2.305 0 011.8-.865 2.305 2.305 0 012.304 2.306 2.305 2.305 0 01-2.305 2.304 2.305 2.305 0 01-1.8-.865v4.226H11.26v-4.258a2.305 2.305 0 001.781.842 2.305 2.305 0 002.305-2.305 2.305 2.305 0 00-2.305-2.305 2.305 2.305 0 00-1.78.841V4.611H7.072a2.305 2.305 0 00.801-1.747A2.305 2.305 0 005.57.559a2.305 2.305 0 00-.009 0z\" opacity=\"").concat(opacity, "\" stroke=\"").concat(stroke, "\" fill=\"").concat(fill, "\" stroke-width=\"").concat(strokeWidth, "\" stroke-linejoin=\"round\"/></svg>");
}; // eslint-disable-line

var initialSizes = {
    WIDTH: 350,
    HEIGHT: 200,
    PUZZLE_SIZE: 60,
    PADDING: 20,
};
var createCaptcha = function (options) { return __awaiter(void 0, void 0, void 0, function () {
    var imageOptions, image, distort, rotate, fill, stroke, strokeWidth, opacity, seed, overlay, mask, outline, location, ins, background, composed, slider;
    var _a, _b, _c, _d;
    return __generator(this, function (_e) {
        switch (_e.label) {
            case 0:
                imageOptions = {
                    width: ((_a = options === null || options === void 0 ? void 0 : options.imageOptions) === null || _a === void 0 ? void 0 : _a.width) || initialSizes.WIDTH,
                    height: ((_b = options === null || options === void 0 ? void 0 : options.imageOptions) === null || _b === void 0 ? void 0 : _b.height) || initialSizes.HEIGHT,
                    puzzleSize: ((_c = options === null || options === void 0 ? void 0 : options.imageOptions) === null || _c === void 0 ? void 0 : _c.puzzleSize) || initialSizes.PUZZLE_SIZE,
                    padding: ((_d = options === null || options === void 0 ? void 0 : options.imageOptions) === null || _d === void 0 ? void 0 : _d.padding) || initialSizes.PADDING,
                };
                image = (options === null || options === void 0 ? void 0 : options.image) ||
                    Buffer.from(backgroundSvg(imageOptions.width, imageOptions.height));
                distort = (options === null || options === void 0 ? void 0 : options.distort) || false;
                rotate = (options === null || options === void 0 ? void 0 : options.rotate) || false;
                fill = (options === null || options === void 0 ? void 0 : options.fill) || "#000";
                stroke = (options === null || options === void 0 ? void 0 : options.stroke) || "#fff";
                strokeWidth = (options === null || options === void 0 ? void 0 : options.strokeWidth) || ".4";
                opacity = (options === null || options === void 0 ? void 0 : options.opacity) || "0.5";
                seed = createRandomInteger();
                overlay = Buffer.from(puzzlePieceSvg({
                    rotate: rotate,
                    distort: distort,
                    fill: fill,
                    stroke: stroke,
                    strokeWidth: strokeWidth,
                    opacity: opacity,
                    seed: seed,
                }));
                mask = Buffer.from(puzzlePieceSvg({
                    rotate: rotate,
                    distort: distort,
                    seed: seed,
                    strokeWidth: strokeWidth,
                    fill: "#fff",
                    stroke: "#fff",
                    opacity: "1",
                }));
                outline = Buffer.from(puzzlePieceSvg({
                    rotate: rotate,
                    distort: distort,
                    seed: seed,
                    stroke: stroke,
                    strokeWidth: strokeWidth,
                    fill: "none",
                    opacity: "1",
                }));
                location = {
                    // Solution for slider
                    left: createRandomInteger(imageOptions.puzzleSize + imageOptions.padding, imageOptions.width - (imageOptions.puzzleSize + imageOptions.padding)),
                    // Vertical offset
                    top: createRandomInteger(imageOptions.padding, imageOptions.height - (imageOptions.puzzleSize + imageOptions.padding)),
                };
                ins = sharp(image).resize({
                    width: imageOptions.width,
                    height: imageOptions.height,
                });
                return [4 /*yield*/, ins
                        .composite([
                        {
                            input: overlay,
                            blend: "over",
                            top: location.top,
                            left: location.left,
                        },
                    ])
                        .png()
                        .toBuffer()];
            case 1:
                background = _e.sent();
                return [4 /*yield*/, ins
                        .composite([
                        {
                            input: mask,
                            blend: "dest-in",
                            top: location.top,
                            left: location.left,
                        },
                        {
                            input: outline,
                            blend: "over",
                            top: location.top,
                            left: location.left,
                        },
                    ])
                        .toBuffer()];
            case 2:
                composed = _e.sent();
                return [4 /*yield*/, sharp(composed)
                        .extract({
                        left: location.left,
                        top: 0,
                        width: imageOptions.puzzleSize,
                        height: imageOptions.height,
                    })
                        .png()
                        .toBuffer()];
            case 3:
                slider = _e.sent();
                return [2 /*return*/, {
                        data: {
                            background: background,
                            slider: slider,
                        },
                        solution: location.left,
                    }];
        }
    });
}); };

// Solution must be correct within the given tolerance
var verifySolution = function (captcha, solution, tolerance) {
    return Math.abs(captcha - solution) < tolerance;
};
// Slider position must not jump to the solution without intermediate values
//TODO: correct any types
var verifyHorizontalMotion = function (positions) {
    return !positions.reduce(function (jumpToInput, pos) {
        return jumpToInput && (pos === 0 || pos === positions[positions.length - 1]);
    }, true);
};
// Vertical motion must be present while dragging the slider
//TODO: correct any types
var verifyVerticalMotion = function (positions) {
    return positions.reduce(function (total, pos) { return total + pos; }) !== 0;
};
var verifyTrailLength = function (trail) { return trail.x.length === trail.y.length; };
var verifyResponse = function (captcha, solution, trail, tolerance) {
    return verifySolution(captcha, solution, tolerance) &&
        verifyTrailLength(trail) &&
        verifyHorizontalMotion((trail.x, solution)) &&
        verifyVerticalMotion(trail.y);
};
//TODO: correct any types
var verifyCaptcha = function (captcha, _a, _b) {
    var response = _a.response, trail = _a.trail;
    var _c = _b === void 0 ? {} : _b, _d = _c.tolerance, tolerance = _d === void 0 ? 7 : _d, _e = _c.verify, verify = _e === void 0 ? verifyResponse : _e;
    return new Promise(function (resolve) {
        if (verify(captcha, response, trail, tolerance)) {
            uid(32).then(function (token) {
                resolve({
                    result: "success",
                    token: token,
                });
            });
        }
        else {
            resolve({ result: "failure" });
        }
    });
};

export { createCaptcha, verifyCaptcha };
